#include "tm.h"
#include "ppc.h"
#include "framebuffer.h"

using namespace std;

#include <fstream>
#include <iostream>

void TM::SetAsAACube(V3 cc, float sideLength) {

	vertsN = 8;
	verts = new V3[vertsN];
	float sl2 = sideLength / 2.0f;
	verts[0] = cc + V3(-sl2, +sl2, +sl2);
	verts[1] = cc + V3(-sl2, -sl2, +sl2);
	verts[2] = cc + V3(+sl2, -sl2, +sl2);
	verts[3] = cc + V3(+sl2, +sl2, +sl2);
	verts[4] = cc + V3(-sl2, +sl2, -sl2);
	verts[5] = cc + V3(-sl2, -sl2, -sl2);
	verts[6] = cc + V3(+sl2, -sl2, -sl2);
	verts[7] = cc + V3(+sl2, +sl2, -sl2);

	trisN = 12;
	tris = new unsigned int[trisN*3];
	int tri = 0;
	tris[3 * tri + 0] = 0; // tris[0] = 0;
	tris[3 * tri + 1] = 1;
	tris[3 * tri + 2] = 2;
	tri++;
	tris[3 * tri + 0] = 2;
	tris[3 * tri + 1] = 3;
	tris[3 * tri + 2] = 0;
	tri++;

	tris[3 * tri + 0] = 3;
	tris[3 * tri + 1] = 2;
	tris[3 * tri + 2] = 6;
	tri++;
	tris[3 * tri + 0] = 6;
	tris[3 * tri + 1] = 7;
	tris[3 * tri + 2] = 3;
	tri++;

	tris[3 * tri + 0] = 7;
	tris[3 * tri + 1] = 6;
	tris[3 * tri + 2] = 5;
	tri++;
	tris[3 * tri + 0] = 5;
	tris[3 * tri + 1] = 4;
	tris[3 * tri + 2] = 7;
	tri++;

	tris[3 * tri + 0] = 4;
	tris[3 * tri + 1] = 5;
	tris[3 * tri + 2] = 1;
	tri++;
	tris[3 * tri + 0] = 1;
	tris[3 * tri + 1] = 0;
	tris[3 * tri + 2] = 4;
	tri++;


	tris[3 * tri + 0] = 1;
	tris[3 * tri + 1] = 5;
	tris[3 * tri + 2] = 6;
	tri++;
	tris[3 * tri + 0] = 6;
	tris[3 * tri + 1] = 2;
	tris[3 * tri + 2] = 1;
	tri++;

	tris[3 * tri + 0] = 4;
	tris[3 * tri + 1] = 0;
	tris[3 * tri + 2] = 3;
	tri++;
	tris[3 * tri + 0] = 3;
	tris[3 * tri + 1] = 7;
	tris[3 * tri + 2] = 4;
	tri++;

	cols = new V3[vertsN];
	cols[0] =
		cols[1] =
		cols[2] =
		cols[3] = V3(1.0f, 0.0f, 0.0f);
	cols[4] =
		cols[5] =
		cols[6] =
		cols[7] = V3(0.0f, 0.0f, 0.0f);
}

#if 0
    4      7

0     3
    5      6
1     2
#endif

void TM::LoadBin(char *fname) {

		ifstream ifs(fname, ios::binary);
		if (ifs.fail()) {
			cerr << "INFO: cannot open file: " << fname << endl;
			return;
		}

		ifs.read((char*)&vertsN, sizeof(int));
		char yn;
		ifs.read(&yn, 1); // always xyz
		if (yn != 'y') {
			cerr << "INTERNAL ERROR: there should always be vertex xyz data" << endl;
			return;
		}
		if (verts)
			delete verts;
		verts = new V3[vertsN];

		ifs.read(&yn, 1); // cols 3 floats
		V3 *normals = 0;
		if (cols)
			delete cols;
		cols = 0;
		if (yn == 'y') {
			cols = new V3[vertsN];
		}

		ifs.read(&yn, 1); // normals 3 floats
		if (normals)
			delete normals;
		normals = 0;
		if (yn == 'y') {
			normals = new V3[vertsN];
		}

		ifs.read(&yn, 1); // texture coordinates 2 floats
		float *tcs = 0; // don't have texture coordinates for now
		if (tcs)
			delete tcs;
		tcs = 0;
		if (yn == 'y') {
			tcs = new float[vertsN * 2];
		}


		ifs.read((char*)verts, vertsN * 3 * sizeof(float)); // load verts

		if (cols) {
			ifs.read((char*)cols, vertsN * 3 * sizeof(float)); // load cols
		}

		if (normals)
			ifs.read((char*)normals, vertsN * 3 * sizeof(float)); // load normals

		if (tcs)
			ifs.read((char*)tcs, vertsN * 2 * sizeof(float)); // load texture coordinates

	ifs.read((char*)&trisN, sizeof(int));
	if (tris)
		delete tris;
	tris = new unsigned int[trisN * 3];
	ifs.read((char*)tris, trisN * 3 * sizeof(unsigned int)); // read tiangles

	ifs.close();

	cerr << "INFO: loaded " << vertsN << " verts, " << trisN << " tris from " << endl << "      " << fname << endl;
	cerr << "      xyz " << ((cols) ? "rgb " : "") << ((normals) ? "nxnynz " : "") << ((tcs) ? "tcstct " : "") << endl;

}


void TM::RenderAsPoints(int psize, PPC *ppc, FrameBuffer *fb) {

	for (int vi = 0; vi < vertsN; vi++) {
		V3 pp;
		if (!ppc->Project(verts[vi], pp))
			continue;
		fb->DrawDisk(pp, (float)psize, 0xFF000000);
	}

}

void TM::RenderWireFrame(PPC *ppc, FrameBuffer *fb) {

	for (int tri = 0; tri < trisN; tri++) {
		for (int ei = 0; ei < 3; ei++) {
			V3 v0 = verts[tris[3 * tri + ei]];
			V3 v1 = verts[tris[3 * tri + (ei + 1) % 3]];
			V3 c0 = cols[tris[3 * tri + ei]];
			V3 c1 = cols[tris[3 * tri + (ei + 1) % 3]];
			V3 pv0, pv1;
			if (!ppc->Project(v0, pv0))
				continue;
			if (!ppc->Project(v1, pv1))
				continue;
			fb->DrawSegment(pv0, pv1, c0, c1);
		}
	}

}

V3 TM::GetCenter() {

	V3 ret(0.0f, 0.0f, 0.0f);
	for (int vi = 0; vi < vertsN; vi++)
		ret = ret + verts[vi];
	ret = ret / (float)vertsN;
	return ret;

}

void TM::Rotate(V3 aO, V3 aD, float theta) {

	for (int vi = 0; vi < vertsN; vi++) {
		verts[vi] = verts[vi].RotateThisPoint(aO, aD, theta);
	}

}


void TM::Translate(V3 tv) {

	for (int vi = 0; vi < vertsN; vi++)
		verts[vi] = verts[vi] + tv;

}

void TM::SetCenter(V3 newCenter) {

	V3 center = GetCenter();
	V3 tv = newCenter - center;
	Translate(tv);

}
